-- EVENTOS

local prefix = 'TriggerRegister'

for name, fun in pairs(_G) do

    if name:sub(1, string.len(prefix)) == prefix then

        new_fun_name = name:sub(string.len(prefix) + 1, string.len(name))
        --print(new_fun_name)

        _G[new_fun_name] = function(...)

            local args = {...}

            --local target = args[1]

            if type(args[1]) == 'table' then
                --target = target.target
                args[1] = args[1].target
            end

            local _callback = table.remove(args) -- eliminamos el ultimo elemento de la tabla -> igual que: table.remove(args, #args)

            local callback = function()
            
                local e = {
                    unit = get_event_unit(),
                    item = get_event_item(),
                    player = get_event_player(),
                    trackable = get_event_trackable()
                }

                return _callback(e)

            end
            
            local trigger = CreateTrigger()

            fun(trigger, table.unpack(args))

            TriggerAddAction(trigger, callback)

            return trigger

        end

    end
    
end


-- UNIDADES

local base_wtables = {}

function mk_wrap_table(target, target_str)

    local Base = nil

    if not base_wtables[target_str] then

        Base = {}
        base_wtables[target_str] = Base

        Base.__index = function(self, attr)

            local method = Base[attr]

            if method ~= nil then

                return method

            elseif method == nil then

                for k, v in pairs(_G) do

                    if k:lower() == attr:lower():gsub('_', '') then

                        method = function(...)
                            return _G[k](self.target, table.unpack({...}))
                        end

                        self[attr] = method

                        return method

                    end

                end

            end

        end

        Base.__eq = function(self, other)
            return self.target == other.target
        end

        Base.__call = function(self)
            return self.target
        end

    else
        Base = base_wtables[target_str]
    end

    local wtable = setmetatable({target = target}, Base)

    for fun_name, callable in pairs(_G) do


        if fun_name:find(target_str) ~= nil and fun_name:find(target_str .. 's') == nil and fun_name:find('Any' .. target_str) == nil and fun_name:find('TriggerRegister') == nil and type(callable) == 'function' then

            --print(fun_name)

            --[[local method_name = fun_name:gsub(target_str, '_'):lower()

            if method_name:sub(-1) == '_' then
                method_name = method_name:sub(1, #method_name-1)
            end]]


            local fixed_fun_name = fun_name:gsub(target_str, '')
            local method_name = ''

            for i = 1, #fixed_fun_name do

                local char = fixed_fun_name:sub(i, i)

                if char == char:upper() then
                    if i > 1 then
                        method_name = method_name .. '_' .. char:lower()
                    else
                        method_name = method_name .. char:lower()
                    end
                else
                    method_name = method_name .. char
                end

            end

            --print(method_name)



            wtable[method_name] = function(...)

                --print(fun_name .. ' -> ' .. method_name)

                local result = callable(wtable.target, table.unpack({...}))

                -- FIX Items >>
                if target_str:lower() == 'unit' then

                    local result_str = tostring(result)

                    local dos_puntos_index = result_str:find(':')

                    if dos_puntos_index and result_str:sub(1, dos_puntos_index-1) == 'item' then

                        result = mk_wrap_table(result, 'Item')

                    end

                end
                -- << FIX Items (para cuando son el resultado de llamar a un metodo de unit)

                return result

            end

        end


    end


    return wtable


end




_G.create_unit = function(...)

    local args = {...}

    if type(args[1]) == 'table' then
        args[1] = args[1].target
    end

    local unit = CreateUnit(table.unpack(args))

    return mk_wrap_table(unit, 'Unit')

end

_G.create_item = function(...)

    local args = {...}

    if type(args[1]) == 'table' then
        args[1] = args[1].target
    end

    local item = CreateItem(table.unpack(args))

    return mk_wrap_table(item, 'Item')

end



_G.create_trackable = function(...)

    local args = {...}

    if type(args[1]) == 'table' then
        args[1] = args[1].target
    end

    local trackable = CreateTrackable(table.unpack(args))

    return mk_wrap_table(trackable, 'Trackable')

end



_G.get_event_unit = function(...)

    local unit = GetTriggerUnit()

    return mk_wrap_table(unit, 'Unit')

end



_G.get_event_item = function(...)

    local item = GetManipulatedItem()

    return mk_wrap_table(item, 'Item')

end



_G.get_event_trackable = function(...)

    local trackable = GetTriggeringTrackable()

    return mk_wrap_table(trackable, 'Trackable')

end



_G.get_event_player = function(...)

    local player = GetTriggerPlayer()

    return mk_wrap_table(player, 'Player')

end


_G.playerlist = setmetatable({}, {
    __index = function(s, i)
        local player = Player(i)
        return mk_wrap_table(player, 'Player')
    end
})


--[[
--Esto mola pero tendría que poner de atributos los nombres exactos en _G, porque buscarlos con capitalize a partir de lower no es facil, simpre habría que iterar _G y comprobar si _G[attr] == _G[attr]:lower()
--Y si quiero quitar la palabra Unit de los métodos, habría que buscar _G[attr] == _G[attr]:gsub('Unit', '_'):lower()
--Pero el problema gordo es que siempre habría que iterar _G para el acceso a cada método si quiero que como método sea lowercase y snakecase...
--Seguramente nada eficiente aunque mole
Unit = {
    __index = function(self, attr)
        
        if attr:lower():find('unit') ~= nil and attr:lower():find('units') == nil then
            return _G[attr]
        end

        return nil

    end
}
--]]


local function mod(a, b)
    return a - math.floor(a / b) * b
end

--[[function _G.reverseFourCC(fourcc)

    local char1 = mod(math.floor(fourcc / 256^3), 256)
    local char2 = mod(math.floor(fourcc / 256^2), 256)
    local char3 = mod(math.floor(fourcc / 256^1), 256)
    local char4 = mod(fourcc, 256)
    
    local result = string.char(char1, char2, char3, char4)

    return result

end]]

function _G.reverseFourCC(fourcc)

    local i = 3

    local ascii_chars = {}

    while i > 0 do

        local char = mod(math.floor(fourcc / 256^i), 256)

        table.insert(ascii_chars, #ascii_chars+1, char)

        i = i - 1

    end

    local char = mod(fourcc, 256)

    table.insert(ascii_chars, #ascii_chars+1, char)

    return string.char(table.unpack(ascii_chars))

end






playerlist[0].chat_event('', false, function(e)


    local codigo = e.player.get_event_chat_string()

    print('>>', codigo)

    local func, err = load(codigo)
    
    if func then
        func()
    else
        print('>> ERROR: ' .. err)
    end

end)
