W3_ENV = (TriggerRegisterTimerEvent ~= nil)


--si estamos en w3
if W3_ENV then
    error = print
end





function hasattr(t, attr)

    local value = rawget(t, attr)

    if value ~= nil then

        return true

    else

        local metatable = getmetatable(t)

        for _, v in ipairs(metatable.nils) do

            if v == attr then
                return true
            end

        end

    end

    return false

end






local MTable = {

    includes = function(self, attr)

        local nils = getmetatable(self).nils

        local value = rawget(self, attr)
    
        if value ~= nil then
    
            return true
    
        else
    
            for _, v in ipairs(nils) do
    
                if v == attr then
                    return true
                end
    
            end
    
        end
    
        return false
    
    end,

    get = function(self, value, default)
    
        return rawget(self, value) or default
    
    end,

    __index = function(self, attr)

        --solo se llega aqui si no se encuentra el attr en la tabla que tiene esta metatabla.
        --tengo que comprobar si attr es nil y en ese caso devolver un error si no está en nils

        local nils = getmetatable(self).nils

        for _, v in ipairs(nils) do
            if v == attr then
                return nil
            end

        end

        local callable = getmetatable(self)[attr] --methods[attr]

        if callable ~= nil then
            return function(...)
                return callable(self, table.unpack({...}))
            end
        end

        error('No existe el atributo \'' .. attr .. '\' en ' .. tostring(self))

    end,

    __newindex = function(self, attr, value)

        local nils = getmetatable(self).nils

        if value == nil then
            nils[#nils+1] = attr

        end

        rawset(self, attr, value) --rawset sería como el setattr de python

    end

}



function Table(t)

    Tab = {}

    for k, v in pairs(MTable) do
        Tab[k] = v
    end

    Tab.nils = {}

    if t == nil then
        t = {}
    end

    for k, v in pairs(t) do

        if v == nil then
            Tab.nils[#Tab.nils+1] = k
        end

    end

    t = setmetatable(t, Tab)

    return t

end





local MList = {

    pop = function(slist, index)
        return table.remove(slist, index)
    end,
    shift = function(slist)
        return MList.pop(slist, 1)
    end,
    insert = function(slist, index, value)
        return table.insert(slist, index, value)
    end,
    append = function(slist, value)
        return table.insert(slist, #slist+1, value)
    end,
    push = function(slist, value)
        return MList.append(slist, value)
    end,
    slice = function(slist, i, f)
    
        -- si incluyo f para ser coherente con los indice 1 y lo que suele suceder en lua

        local new_list = list()

        for index = i, f do

            new_list.append(slist[index])

        end

        return new_list

    end,

    __index = function(self, index)

        if type(index) ~= 'number' then

            local callable = getmetatable(self)[index] --methods[index] -- aqui index es un attr, no un numero

            if callable ~= nil then
                return function(...)
                    return callable(self, table.unpack({...}))
                end
            end

        end


        if type(index) == 'number' and index == math.floor(index) and index <= #self+1 then
            return rawget(self, index)
        else
            error('ERROR __index: \'' .. tostring(index) .. '\' no es un indice valido para la lista ' .. tostring(self))
        end
    
    end,
    
    __newindex = function(self, index, value)

        local indices = getmetatable(self).indices

        if type(index) == 'number' and index == math.floor(index) and #indices <= index then
            rawset(self, index, value)
            indices[#indices+1] = index
        else
            error('ERROR __newindex: \'' .. tostring(index) .. '\' no es un indice valido para la lista ' .. tostring(self))
        end

    end,

    __add = function(self, other)
    
        local new_list = list()

        for _, val in ipairs(self) do
            new_list.append(val)
        end

        for _, val in ipairs(other) do
            new_list.append(val)
        end

        return new_list
    
    end,

    __mul = function(self, n)
    
        local new_list = list()

        for i = 1, n do
            new_list = new_list + self
        end

        return new_list
    
    end

}



function list(...)

    local List = {}

    for k, v in pairs(MList) do
        List[k] = v
    end

    List.indices = {}

    local t = {...}

    if t == nil then
        t = {}
    end

    for i, v in ipairs(t) do
        List.indices[i] = v
    end

    return setmetatable(t, List)


end




if _VERSION == 'Lua 5.3' and hasattr(_G, 'CreateTimer') then

    --para no afectar a lo que hace blizzard
    local timer = CreateTimer()
    TimerStart(timer, 0, false, function()

        --print(getmetatable(_G))
        Table(_G)
        --print(getmetatable(_G))

    end)

end



if not W3_ENV then

    --print(error == print)

    local d = Table({
        dude=111,
        otro=222
    })

    --print(d.get('dude'))
    if d.includes('otro') then
        print(d.otro)
    end


    local lista = list({'b', 'c'})

    lista.append('d')
    lista.insert(1, 'a')

    for index, value in ipairs(lista) do
        print(value)
    end


    print('||||||||')

    for i, v in ipairs(list(1,2,3)) do print(v) end



end